from __future__ import division
import numpy as np
import scipy.fftpack as scipyfft
import threading
import pyaudio
import aubio
import time

class AudioStream():
    def __init__(self):
        self.CHANNELS = 1
        self.RATE = 44100
        self.CHUNK = 1024*4
        self.frames_per_buffers = int(256)

    # Should return a numpy float32 array containing the audio data
    def update(self):
        return []

class AudioStreamPyAudio(AudioStream):
    def __init__(self, RATE=44100, device=0):
        AudioStream.__init__(self)
        
        self.RATE = RATE

        self.p = pyaudio.PyAudio()

        self.stream = self.p.open(
            format=pyaudio.paInt16,
            channels=self.CHANNELS,
            input_device_index=device,
            rate=self.RATE,
            input=True,
            output=True,
            frames_per_buffer=self.frames_per_buffers,
        )

    def update(self):
        data = self.stream.read(self.frames_per_buffers, exception_on_overflow=False)
        self.stream.read(self.stream.get_read_available(), exception_on_overflow=False)
        data = np.fromstring(data, dtype=np.int16)
        return data.astype(np.float32)

class AudioHandler(threading.Thread):

    def __init__(self, width, height, audio_stream):
        threading.Thread.__init__(self)

        self.dimensions = [width, height]
        self.audio = audio_stream

        self.bassMax = 0

        self.lastFFTsSize = 4
        self.lastFFTs = [[] for _ in range(self.lastFFTsSize)]
        self.lastFFTsCounter = 0

        self.listeners = []

        self.pv = aubio.pvoc(self.audio.frames_per_buffers*2, self.audio.frames_per_buffers)
        self.onset = aubio.onset("specdiff", self.audio.frames_per_buffers*2, self.audio.frames_per_buffers, self.audio.RATE)
        self.tempo = aubio.tempo("default", self.audio.frames_per_buffers*2, self.audio.frames_per_buffers, self.audio.RATE)

        self.weight_filter = aubio.digital_filter(7)
        self.weight_filter.set_a_weighting(self.audio.RATE)

        self.beat = 0
        self.beats = []

        self.time_started = -1

        self.pause_cond = threading.Condition(threading.Lock())
        self.paused = False

    def registerListener(self, listener):
        self.listeners.append(listener)

    def removeListener(self, listener):
        self.listeners.remove(listener)

    def beats_to_bpm(self):
        # if enough beats are found, convert to periods then to bpm
        if len(self.beats) > 1:
            bpms = 60./np.diff(self.beats)
            return np.median(bpms)
        return 0
    def pause(self):
        self.paused = True
        self.pause_cond.acquire()
    def resume(self):
        self.paused = False
        self.pause_cond.notify()
        self.pause_cond.release()

    def run(self):
        while True:
            with self.pause_cond:
                while self.paused:
                    self.pause_cond.wait()

                data = self.audio.update()

                if self.time_started == -1:
                    self.time_started = time.time()

                #sp_data = np.abs(scipyfft.rfft(data))
                sp_data = self.pv(self.weight_filter(data)).norm

                # Normalisation 
                sp_amp = []
                norm = 0
                total = 0
                for i in range(len(sp_data)):
                    value = np.sqrt(np.power(np.real(sp_data[i]), 2) + np.power(np.imag(sp_data[i]), 2))
                    sp_amp.append(value)
                    norm += value
                    total += 1

                sp_amp /= norm/total

                maximum = -1
                for i in range(len(sp_amp)):
                    if maximum == -1 or sp_amp[i] > maximum:
                        maximum = sp_amp[i]
                sp_amp /= maximum*0.8
                if np.isfinite(sp_amp).all():

                    # Average
                    self.lastFFTs[self.lastFFTsCounter] = sp_amp
                    self.lastFFTsCounter += 1
                    if (self.lastFFTsCounter >= self.lastFFTsSize):
                        self.lastFFTsCounter = 0


                    # Subdivide the frequency spectrum logarithmically because the interesting frequencies are not linearly spaced
                    #frequencies = np.geomspace(125, self.audio.RATE//2, 1+self.dimensions[0]) # width + 1 because we want a special bass channel < 125Hz
                    frequencies = [64, 125, 250, 500, 1000, 2000, 4000, 8000, 16000]
                    # Depending on the sample rate, we get different frequency steps that we must account for
                    sampled_freqs = np.linspace(0, self.audio.RATE//2, len(sp_amp))

                    # The final table with each frequency group averaged out
                    averaged_out = [0 for _ in range(len(frequencies))]

                    (data_counter, freq_counter, start) = (0, 0, 0) # Loop counters
                    while data_counter < len(sp_amp):
                        value = sum([(self.lastFFTs[k][data_counter] if data_counter < len(self.lastFFTs[k]) else 0) for k in range(self.lastFFTsSize)]) / self.lastFFTsSize
                        #value *= (1-0.9*np.exp(-3*i/len(sp_amp))) -- TODO ? Human hearing compensation
                        #print(value)
                        if freq_counter < len(frequencies):
                            #print(value)
                            #if value <= 0.00001:
                            #    value = 0
                            if sampled_freqs[data_counter] >= frequencies[freq_counter]:
                                averaged_out[freq_counter] /= data_counter - start
                                start = data_counter
                                freq_counter += 1
                                continue
                            averaged_out[freq_counter] += value
                        data_counter += 1
                    #averaged_out[freq_counter] /= data_counter - start

                    # Separate the bass from the rest of the table
                    bass = averaged_out[0]
                    if bass > self.bassMax:
                        self.bassMax = bass*0.9
                    averaged_out = averaged_out[1:]
                    average = sum(averaged_out)/len(averaged_out)

                    onsetDetected = self.onset(data)

                    beatDetected = self.tempo(data)
                    if beatDetected:
                        self.beats.append(self.tempo.get_last_s())
                    
                    maximum = 0
                    for i in averaged_out:
                        if i > maximum:
                            maximum = i
                    # Notify the listeners of a change
                    for listener in self.listeners:
                        listener.onAudioUpdate(averaged_out, maximum, 1 if onsetDetected else 0)
                        listener.onBPMUpdate(self.beats_to_bpm(), beatDetected, self.time_started + self.tempo.get_last_s())
