from __future__ import division
from processing import DefaultHandler

class BarDisplayHandler(DefaultHandler):
    class Segment:
        ANIMATION_TABLE = {
            "fade_in": lambda value, progress, duration, data: value,
            "fade_out": lambda value, progress, duration, data: (1-progress/duration)*data[0]
        }

        ANIMATION_IDS = {
            "fade_in": 0,
            "fade_out": 1
        }

        ID_TO_ANIMATION = ["fade_in", "fade_out"]

        def __init__(self, height, frequencyGroup):
            self.height = height
            self.frequencyGroup = frequencyGroup

            self.value = 0
            self.prevValue = 0

            self.animationTimer = 0
            self.activeAnimations = [False for _ in range(len(self.ANIMATION_IDS))]

        def update(self, value, nextSegment):
            self.prevValue = self.value
            self.value = value

            if self.hasChanged() == 1: # This segment just became active
                if self.hasAnimation("fade_out"):
                    self.stopAnimation("fade_out")
                if not self.hasAnimation("fade_in"):
                    self.startAnimation("fade_in", 100, [0.4])
            elif self.hasChanged() == -1:
                if not self.hasAnimation("fade_out"):
                    if self.hasAnimation("fade_in"):
                        pass
                        #self.stopAnimation("fade_in")
                    else:
                        pass
                        #self.startAnimation("fade_out", 20, [self.prevValue])

            if nextSegment != None and nextSegment.getValue() > 0 and self.hasAnimation("fade_out"): # Prevent a fadeout if we're not the last segment
                self.value = self.activeAnimations[self.ANIMATION_IDS["fade_out"]][2][0]
                self.stopAnimation("fade_out")

            self.animationTimer += 1

            activeAnimations = self.activeAnimations
            for i in range(len(self.activeAnimations)):
                if not self.activeAnimations[i]:
                    continue
                animation = self.ID_TO_ANIMATION[i]
                (start, duration, data) = self.activeAnimations[i]

                if (self.animationTimer - start <= duration):
                    self.value = self.ANIMATION_TABLE[animation](self.value, self.animationTimer - start, duration, data)
                else:
                    activeAnimations[i] = False
            self.activeAnimations = activeAnimations

        def stopAnimation(self, name):
            self.activeAnimations[self.ANIMATION_IDS[name]] = False
        def startAnimation(self, name, duration, data):
            self.activeAnimations[self.ANIMATION_IDS[name]] = (self.animationTimer, duration, data)

        def getAnimationProgress(self, name):
            if not self.hasAnimation(name):
                return 0
            (start, duration, data) = self.activeAnimations[self.ANIMATION_IDS[name]]
            return (self.animationTimer - start) / duration
        def getAnimationData(self, name):
            if not self.hasAnimation(name):
                return []
            (start, duration, data) = self.activeAnimations[self.ANIMATION_IDS[name]]
            return data
        def hasAnimationID(self, anim_id):
            return True if self.activeAnimations[anim_id] else False
        def hasAnimation(self, name):
            return self.hasAnimationID(self.ANIMATION_IDS[name])

        def getValue(self):
            return self.value
        def isActive(self):
            return self.value != 0
        def wasActive(self):
            return self.prevValue != 0
        def hasChanged(self):
            direction = 0
            if self.isActive() and not self.wasActive():
                direction = 1
            elif not self.isActive() and self.wasActive():
                direction = -1
            return direction
    class Bar:
        def __init__(self, index, height):
            self.index = index
            self.segments = [BarDisplayHandler.Segment(i, self) for i in range(height)]

            self.max = 0
            self.value = 0

            self.prev10Values = [0 for _ in range(10)]
            self.prev10ValuesCounter = 0

            self.averageChangeRateSum = 0
            self.averageChangeRateCount = 0

        def update(self, value, global_average):
            if (value > self.max):
                self.max = value

            self.prev10ValuesCounter += 1
            if (self.prev10ValuesCounter >= 10):
                self.prev10ValuesCounter = 0
            self.prev10Values[self.prev10ValuesCounter] = value

            self.value = value

            self.averageChangeRateSum += self.getScaledDelta()
            self.averageChangeRateCount += 1
            if (self.averageChangeRateCount > 1024):
                self.averageChangeRateCount /= 2
                self.averageChangeRateSum /= 2

            intValue = self.getINTValue(global_average, len(self.segments))
            for j in range(intValue):
                self.segments[j].update(self.value, self.segments[j+1] if j+1 < len(self.segments) else None)
            for j in range(intValue, len(self.segments)):
                self.segments[j].update(0, self.segments[j+1] if j+1 < len(self.segments) else None)

            return self.value

        def getLocalAverage(self):
            return sum(self.prev10Values)/10

        def getAverageChangeRate(self):
            return self.averageChangeRateSum/self.averageChangeRateCount

        def getScaledDelta(self):
            prevValue = sum(self.prev10Values)/10
            return 0 if prevValue == 0 else max(0, (self.value - prevValue) / prevValue)

        def getValue(self):
            return self.value

        def getScaledValue(self):
            return self.value*10*(self.getAverageChangeRate()-0.1)


        def getINTValue(self, global_max, count): # Conversion to INTeger
            return max(0, min(len(self.segments), int(self.value*count/global_max))) if global_max != 0 else 0


    def __init__(self, width, height):
        DefaultHandler.__init__(self, width, height)

        self.channels = [self.Bar(i, self.height) for i in range(self.width)]

        self.values = [[[-1,-1,-1] for j in range(len(self.channels[i].segments))] for i in range(len(self.channels))]
        self.prevValues = [[[-1,-1,-1] for j in range(len(self.channels[i].segments))] for i in range(len(self.channels))]
        self.changed_values = []

    def update(self, data):
        for i in range(len(self.channels)):
            self.channels[i].update(data[i], self.average)

        self.prevValues = self.values
        self.values = [[[-1,-1,-1] for j in range(len(self.channels[i].segments))] for i in range(len(self.channels))]
        self.changed_values = []
        for i in range(len(self.channels)):
            for j in range(len(self.channels[i].segments)):
                # saturation : 0.8 + self.channels[i].getScaledDelta()*0.5 ?
                if (self.bass != 0):
                    self.channels[i].segments[j].startAnimation("fade_in", 10, [0.6])
                value = self.channels[i].segments[j].getValue()/(1 if self.channels[i].getLocalAverage() == 0 else self.channels[i].getLocalAverage())
                if value > 1:
                    value = 1
                    """max(0, min(1-self.channels[i].getScaledDelta(), 1)) if value != 0 else 0"""
                colours = [(0.45 - (j+1)/self.height*0.5), self.channels[i].segments[j].getAnimationData("fade_in")[0] if self.channels[i].segments[j].hasAnimation("fade_in") else 1-self.bass,  0.8 if value != 0 else 0]
                self.changed_values.append([i, j, colours])
                self.values[i][j] = colours

    def getValues(self):
        return (self.changed_values, "hsv")
