from __future__ import division
from ledmatrix import LEDMatrix, LED

import threading
import time
import numpy as np

class DefaultHandler:

    def __init__(self, width, height):  
        self.average = 0
        self.bass = 0

        self.width = width
        self.height = height

    def update(self, data):
        pass
    def setBass(self, bass):
        self.bass = bass
    def setAverage(self, average):
        self.average = average

    def getValues(self):
        return ([[[1,0,1] for j in range(self.height)] for i in range(self.width)], "hsv")

from bardisplay_handler import BarDisplayHandler
from text_handler import TextHandler

class Processing(threading.Thread):

    MODES = ["default", "bar_display", "text"]
    MODES_HANDLERS = {
        "default": DefaultHandler,
        "bar_display": BarDisplayHandler,
        "text": TextHandler
    }

    def __init__(self, led_matrix, refresh_rate):
        threading.Thread.__init__(self)

        self.led_matrix = led_matrix
        self.refresh_rate = refresh_rate

        self.bass = 0

        self.pulse = 0
        self.phase_shift = 0

        self.active_modes = ["bar_display", "text"]
        self.modes = []
        for mode in self.active_modes:
            self.modes.append(self.MODES_HANDLERS[mode](self.led_matrix.width, self.led_matrix.height))

        self.lastRun = time.time()

        self.shouldStop = False
        self.pause_cond = threading.Condition(threading.Lock())
        self.paused = False

    def stop(self):
        self.shouldStop = True
    def pause(self):
        self.paused = True
        self.pause_cond.acquire()
    def resume(self):
        self.paused = False
        self.pause_cond.notify()
        self.pause_cond.release()

    def onAudioUpdate(self, data, average, bass):
        for mode in self.modes:
            mode.setAverage(average)
            mode.setBass(bass)
            self.bass = bass
            mode.update(data)

    def onBPMUpdate(self, bpm, has_beat, last_beat):
        self.pulse = np.pi*bpm/60
        if has_beat:
            self.phase_shift = np.pi/2 - self.pulse*last_beat

    def brightness(self, t):
        phase = self.pulse*t + self.phase_shift
        return np.abs(np.sin(phase))

    def run(self):
        check_colour_mode = lambda colour_mode, final: True if final == "" or final == colour_mode else False

        while not self.shouldStop:
            with self.pause_cond:
                while self.paused:
                    self.pause_cond.wait()
                self.lastRun = time.time()
                final_colour_mode = ""

                length = len(self.modes)

                # Process everything in reverse order because the first mode is the most "invasive"
                for k in range(length):
                    i = length - 1 - k

                    (changing_leds, colour_mode) = self.modes[i].getValues()

                    if check_colour_mode(colour_mode, final_colour_mode):
                        final_colour_mode = colour_mode
                    else:
                        continue

                    for led_status in changing_leds:
                        led = self.led_matrix.getLED(led_status[0], led_status[1])
                            
                        if k == 0:
                            led.setPreparing(True)
                            led.setColourAuto(colour_mode, *led_status[2])
                            led.setBrightness(4)
                        else: # apply blending
                            led_colour = list(led.getColourHSV()) # TODO, handle not-HSV

                            if led_colour[2] != 0 and led.isPreparing(): # blending
                                pass
                                #for l in range(len(led_colour)):
                                #    led_colour[l] = (1-led_status[2][2])*led_colour[l] + led_status[2][2]*led_status[2][l]
                            else: # no blending
                                led_colour = led_status[2]
                                led.setBrightness(1)
                            led.setPreparing(True)
                            led.setColourAuto(colour_mode, *led_colour)                         

                # Send the final update all at once to avoid random flicker
                for led in self.led_matrix.iterator():
                    led[2].send()

                self.led_matrix.brightness = self.brightness(time.time())
                time.sleep(1/self.refresh_rate)
